package dev.fyloz.trial.colorrecipesexplorer.core.configuration

import com.fasterxml.jackson.module.kotlin.jacksonObjectMapper
import dev.fyloz.trial.colorrecipesexplorer.core.model.Employee
import dev.fyloz.trial.colorrecipesexplorer.core.model.EmployeeLoginRequest
import dev.fyloz.trial.colorrecipesexplorer.core.model.EmployeePermission
import dev.fyloz.trial.colorrecipesexplorer.core.services.EmployeeService
import dev.fyloz.trial.colorrecipesexplorer.core.services.EmployeeUserDetailsService
import io.jsonwebtoken.Jwts
import io.jsonwebtoken.SignatureAlgorithm
import org.slf4j.Logger
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.boot.context.properties.ConfigurationProperties
import org.springframework.boot.context.properties.EnableConfigurationProperties
import org.springframework.context.annotation.Bean
import org.springframework.context.annotation.Configuration
import org.springframework.http.HttpMethod
import org.springframework.security.authentication.AuthenticationManager
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken
import org.springframework.security.config.annotation.authentication.builders.AuthenticationManagerBuilder
import org.springframework.security.config.annotation.method.configuration.EnableGlobalMethodSecurity
import org.springframework.security.config.annotation.web.builders.HttpSecurity
import org.springframework.security.config.annotation.web.configuration.EnableWebSecurity
import org.springframework.security.config.annotation.web.configuration.WebSecurityConfigurerAdapter
import org.springframework.security.config.annotation.web.configurers.ExpressionUrlAuthorizationConfigurer
import org.springframework.security.config.http.SessionCreationPolicy
import org.springframework.security.core.Authentication
import org.springframework.security.core.AuthenticationException
import org.springframework.security.core.context.SecurityContextHolder
import org.springframework.security.core.userdetails.User
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.security.web.AuthenticationEntryPoint
import org.springframework.security.web.authentication.UsernamePasswordAuthenticationFilter
import org.springframework.security.web.authentication.www.BasicAuthenticationFilter
import org.springframework.stereotype.Component
import org.springframework.util.Assert
import org.springframework.web.cors.CorsConfiguration
import org.springframework.web.cors.CorsConfigurationSource
import org.springframework.web.cors.UrlBasedCorsConfigurationSource
import java.util.*
import javax.annotation.PostConstruct
import javax.servlet.FilterChain
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse

@Configuration
@EnableWebSecurity
@EnableGlobalMethodSecurity(prePostEnabled = true)
@EnableConfigurationProperties(SecurityConfigurationProperties::class)
class WebSecurityConfig(
        val restAuthenticationEntryPoint: RestAuthenticationEntryPoint,
        val securityConfigurationProperties: SecurityConfigurationProperties,
        val logger: Logger) : WebSecurityConfigurerAdapter() {
    @Autowired
    private lateinit var userDetailsService: EmployeeUserDetailsService

    @Autowired
    private lateinit var employeeService: EmployeeService

    override fun configure(authBuilder: AuthenticationManagerBuilder) {
        authBuilder.userDetailsService(userDetailsService).passwordEncoder(passwordEncoder())
    }

    @Bean
    fun passwordEncoder(): PasswordEncoder {
        return BCryptPasswordEncoder()
    }

    @Bean
    override fun authenticationManagerBean(): AuthenticationManager {
        return super.authenticationManagerBean()
    }

    @Bean
    fun corsConfigurationSource(): CorsConfigurationSource {
        return UrlBasedCorsConfigurationSource().apply {
            registerCorsConfiguration("/**", CorsConfiguration().applyPermitDefaultValues())
        }
    }

    @PostConstruct
    fun createRootUser() {
        val rootUserCredentials = securityConfigurationProperties.root
        Assert.notNull(rootUserCredentials, "No root user has been defined.")
        Assert.notNull(rootUserCredentials!!.id, "The root user has no identifier defined.")
        Assert.notNull(rootUserCredentials.password, "The root user has no password defined.")
        if (!employeeService.existsById(rootUserCredentials.id!!)) {
            employeeService.save(Employee(rootUserCredentials.id!!, "Root", "Employee", passwordEncoder().encode(rootUserCredentials.password!!), true, permissions = mutableListOf(EmployeePermission.ADMIN)))
        }
    }

    override fun configure(http: HttpSecurity) {
        fun ExpressionUrlAuthorizationConfigurer<HttpSecurity>.ExpressionInterceptUrlRegistry.generateAuthorizations():
                ExpressionUrlAuthorizationConfigurer<HttpSecurity>.ExpressionInterceptUrlRegistry {
            ControllerAuthorizations.values().forEach { controller ->
                val antMatcher = controller.antMatcher
                controller.permissions.forEach {
                    antMatchers(it.key, antMatcher).hasAuthority(it.value.name)
                    logger.debug("Added authorization for path '$antMatcher': ${it.key.name} -> ${it.value.name}")
                }
            }
            return this
        }

        http
                .cors()
                .and()
                .csrf().disable()
                .authorizeRequests()
                .antMatchers(HttpMethod.GET, "/").permitAll()
                .antMatchers("/api/login").permitAll()
                .antMatchers(HttpMethod.GET, "/api/employee/current").authenticated()
                .generateAuthorizations()
                .and()
                .addFilter(JwtAuthenticationFilter(authenticationManager(), employeeService, securityConfigurationProperties))
                .addFilter(JwtAuthorizationFilter(userDetailsService, securityConfigurationProperties, authenticationManager()))
                .sessionManagement().sessionCreationPolicy(SessionCreationPolicy.STATELESS)
    }
}

@Component
class RestAuthenticationEntryPoint : AuthenticationEntryPoint {
    override fun commence(request: HttpServletRequest, response: HttpServletResponse, authException: AuthenticationException) = response.sendError(HttpServletResponse.SC_UNAUTHORIZED, "Unauthorized")
}

class JwtAuthenticationFilter(
        val authManager: AuthenticationManager,
        val employeeService: EmployeeService,
        val securityConfigurationProperties: SecurityConfigurationProperties
) : UsernamePasswordAuthenticationFilter() {
    init {
        setFilterProcessesUrl("/api/login")
    }

    override fun attemptAuthentication(request: HttpServletRequest, response: HttpServletResponse): Authentication {
        val loginRequest = jacksonObjectMapper().readValue(request.inputStream, EmployeeLoginRequest::class.java)
        return authManager.authenticate(UsernamePasswordAuthenticationToken(loginRequest.id, loginRequest.password))
    }

    override fun successfulAuthentication(request: HttpServletRequest, response: HttpServletResponse, chain: FilterChain, authResult: Authentication) {
        val jwtSecret = securityConfigurationProperties.jwtSecret
        val jwtDuration = securityConfigurationProperties.jwtDuration
        Assert.notNull(jwtSecret, "No JWT secret has been defined.")
        Assert.notNull(jwtDuration, "No JWT duration has been defined.")
        val employeeId = (authResult.principal as User).username
        employeeService.updateLastLoginTime(employeeId.toLong())
        val token = Jwts.builder()
                .setSubject(employeeId)
                .setExpiration(Date(System.currentTimeMillis() + jwtDuration!!))
                .signWith(SignatureAlgorithm.HS512, jwtSecret!!.toByteArray())
                .compact()
        response.addHeader("Authorization", "Bearer $token")
    }
}

class JwtAuthorizationFilter(
        val userDetailsService: EmployeeUserDetailsService,
        val securityConfigurationProperties: SecurityConfigurationProperties,
        authenticationManager: AuthenticationManager
) : BasicAuthenticationFilter(authenticationManager) {
    override fun doFilterInternal(request: HttpServletRequest, response: HttpServletResponse, chain: FilterChain) {
        val header = request.getHeader("Authorization")
        if (header != null && header.startsWith("Bearer")) {
            val authenticationToken = getAuthentication(request)
            SecurityContextHolder.getContext().authentication = authenticationToken
        }
        chain.doFilter(request, response)
    }

    private fun getAuthentication(request: HttpServletRequest): UsernamePasswordAuthenticationToken? {
        val jwtSecret = securityConfigurationProperties.jwtSecret
        Assert.notNull(jwtSecret, "No JWT secret has been defined.")
        val token = request.getHeader("Authorization")
        if (token != null) {
            val employeeId = Jwts.parser()
                    .setSigningKey(jwtSecret!!.toByteArray())
                    .parseClaimsJws(token.replace("Bearer", ""))
                    .body
                    .subject
            return if (employeeId != null) {
                val employeeDetails = userDetailsService.loadUserByUsername(employeeId)
                UsernamePasswordAuthenticationToken(employeeDetails.username, null, employeeDetails.authorities)
            } else null
        }
        return null
    }
}

@ConfigurationProperties("cre.security")
class SecurityConfigurationProperties {
    var jwtSecret: String? = null
    var jwtDuration: Long? = null
    var root: RootUserCredentials? = null

    class RootUserCredentials(var id: Long? = null, var password: String? = null)
}

private enum class ControllerAuthorizations(
        val antMatcher: String,
        val permissions: Map<HttpMethod, EmployeePermission>
) {
    EMPLOYEE_GROUP("/api/employee/group/**", mapOf(
            HttpMethod.GET to EmployeePermission.VIEW_EMPLOYEE_GROUP,
            HttpMethod.POST to EmployeePermission.EDIT_EMPLOYEE_GROUP,
            HttpMethod.PUT to EmployeePermission.EDIT_EMPLOYEE_GROUP,
            HttpMethod.DELETE to EmployeePermission.REMOVE_EMPLOYEE_GROUP
    )),
    EMPLOYEE("/api/employee/**", mapOf(
            HttpMethod.GET to EmployeePermission.VIEW_EMPLOYEE,
            HttpMethod.POST to EmployeePermission.EDIT_EMPLOYEE,
            HttpMethod.PUT to EmployeePermission.EDIT_EMPLOYEE,
            HttpMethod.DELETE to EmployeePermission.REMOVE_EMPLOYEE
    ))
}
