package dev.fyloz.trial.colorrecipesexplorer.core.model

import com.fasterxml.jackson.annotation.JsonIgnore
import org.hibernate.annotations.Fetch
import org.hibernate.annotations.FetchMode
import org.springframework.security.core.GrantedAuthority
import org.springframework.security.core.authority.SimpleGrantedAuthority
import java.time.LocalDateTime
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.NotNull
import javax.validation.constraints.Size


private const val EMPLOYEE_ID_NULL_MESSAGE = "Un numéro d'employé est requis"
private const val EMPLOYEE_LAST_NAME_EMPTY_MESSAGE = "Un nom  est requis"
private const val EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE = "Un prénom est requis"
private const val EMPLOYEE_PASSWORD_EMPTY_MESSAGE = "Un mot de passe est requis"
private const val EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE = "Le mot de passe doit contenir au moins 8 caractères"

@Entity
class Employee(
        @Id
        @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
        override val id: Long,

        val firstName: String = "",

        val lastName: String = "",

        @JsonIgnore
        val password: String = "",

        @JsonIgnore
        val isRoot: Boolean = false,

        @field:ManyToOne
        @Fetch(FetchMode.SELECT)
        var group: EmployeeGroup? = null,

        @Enumerated(EnumType.STRING)
        @ElementCollection(fetch = FetchType.EAGER)
        val permissions: MutableList<EmployeePermission> = mutableListOf(),

        @Enumerated(EnumType.STRING)
        @ElementCollection(fetch = FetchType.EAGER)
        @Fetch(FetchMode.SUBSELECT)
        val excludedPermissions: MutableList<EmployeePermission> = mutableListOf(),

        val lastLoginTime: LocalDateTime? = null
) : IModel

/** DTO for creating employees. The [Employee] entity doesn't allow to modify passwords. */
data class EmployeeDto(
        @field:NotNull(message = EMPLOYEE_ID_NULL_MESSAGE)
        val id: Long,

        @field:NotBlank(message = EMPLOYEE_FIRST_NAME_EMPTY_MESSAGE)
        val firstName: String,

        @field:NotBlank(message = EMPLOYEE_LAST_NAME_EMPTY_MESSAGE)
        val lastName: String,

        @field:NotBlank(message = EMPLOYEE_PASSWORD_EMPTY_MESSAGE)
        @field:Size(min = 8, message = EMPLOYEE_PASSWORD_TOO_SHORT_MESSAGE)
        val password: String,

        @field:ManyToOne
        @Fetch(FetchMode.SELECT)
        var group: EmployeeGroup? = null,

        @Enumerated(EnumType.STRING)
        @ElementCollection(fetch = FetchType.EAGER)
        val permissions: MutableList<EmployeePermission> = mutableListOf(),

        @Enumerated(EnumType.STRING)
        @ElementCollection(fetch = FetchType.EAGER)
        @Fetch(FetchMode.SUBSELECT)
        val excludedPermissions: MutableList<EmployeePermission> = mutableListOf()
)

private const val GROUP_NAME_NULL_MESSAGE = "Un nom est requis"
private const val GROUP_PERMISSIONS_EMPTY_MESSAGE = "Au moins une permission est requise"

@Entity
data class EmployeeGroup(
        @Id
        @GeneratedValue(strategy = GenerationType.SEQUENCE)
        override val id: Long? = null,

        @Column(unique = true)
        @field:NotBlank(message = GROUP_NAME_NULL_MESSAGE)
        @field:Size(min = 3)
        val name: String = "",

        @Enumerated(EnumType.STRING)
        @ElementCollection(fetch = FetchType.EAGER)
        @field:Size(min = 1, message = GROUP_PERMISSIONS_EMPTY_MESSAGE)
        val permissions: MutableList<EmployeePermission> = mutableListOf(),

        @OneToMany
        @JsonIgnore
        val employees: MutableList<Employee> = mutableListOf()
) : IModel


data class EmployeeLoginRequest(val id: Long, val password: String)


enum class EmployeePermission(val impliedPermissions: List<EmployeePermission> = listOf()) {
    // View
    VIEW_EMPLOYEE,
    VIEW_EMPLOYEE_GROUP,
    VIEW(listOf(
    )),

    // Edit
    EDIT_EMPLOYEE,
    EDIT_EMPLOYEE_GROUP,
    EDIT(listOf(
    )),

    // Remove
    REMOVE_EMPLOYEE,
    REMOVE_EMPLOYEE_GROUP,
    REMOVE(listOf(
    )),

    ADMIN(listOf(
            VIEW,
            EDIT,
            REMOVE,

            // Admin only permissions
            VIEW_EMPLOYEE,
            VIEW_EMPLOYEE_GROUP,
            EDIT_EMPLOYEE,
            EDIT_EMPLOYEE_GROUP,
            REMOVE_EMPLOYEE,
            REMOVE_EMPLOYEE_GROUP
    ));

    operator fun contains(permission: EmployeePermission): Boolean {
        return permission == this || impliedPermissions.any { permission in it }
    }
}

/** Gets [GrantedAuthority]s of the given [Employee]. */
fun Employee.getAuthorities(): MutableCollection<GrantedAuthority> {
    return getPermissions().map { it.toAuthority() }.toMutableSet()
}

/** Gets [EmployeePermission]s of the given [Employee]. */
fun Employee.getPermissions(): Iterable<EmployeePermission> {
    val grantedPermissions: MutableSet<EmployeePermission> = mutableSetOf()
    if (group != null) grantedPermissions.addAll(group!!.permissions.flatMap { it.flat() }.filter { excludedPermissions.isEmpty() || excludedPermissions.any { excludedPermission -> it !in excludedPermission } })
    grantedPermissions.addAll(permissions.flatMap { it.flat() }.filter { excludedPermissions.isEmpty() || excludedPermissions.any { excludedPermission -> it !in excludedPermission } })
    return grantedPermissions
}

private fun EmployeePermission.flat(): Iterable<EmployeePermission> {
    return mutableSetOf(this).apply {
        impliedPermissions.forEach {
            addAll(it.flat())
        }
    }
}

/** Converts the given [EmployeePermission] to a [GrantedAuthority]. */
private fun EmployeePermission.toAuthority(): GrantedAuthority {
    return SimpleGrantedAuthority(name)
}
