package dev.fyloz.trial.colorrecipesexplorer.core.services

import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.*
import dev.fyloz.trial.colorrecipesexplorer.core.model.*
import dev.fyloz.trial.colorrecipesexplorer.dao.EmployeeGroupRepository
import dev.fyloz.trial.colorrecipesexplorer.dao.EmployeeRepository
import org.springframework.security.core.userdetails.User
import org.springframework.security.core.userdetails.UserDetails
import org.springframework.security.core.userdetails.UserDetailsService
import org.springframework.security.core.userdetails.UsernameNotFoundException
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.stereotype.Service
import java.time.LocalDateTime
import javax.transaction.Transactional

@Service
class EmployeeService(val employeeRepository: EmployeeRepository, val passwordEncoder: PasswordEncoder) :
        AbstractModelService<Employee, EmployeeRepository>(employeeRepository, Employee::class.java) {
    /** Check if an [Employee] with the given [firstName] and [lastName] exists. */
    fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean {
        return repository.existsByFirstNameAndLastName(firstName, lastName)
    }

    override fun getAll(): Collection<Employee> {
        return super.getAll().filter { !it.isRoot }
    }

    override fun getById(id: Long): Employee {
        return getById(id, true)
    }

    /** Gets the employee with the given [id]. */
    fun getById(id: Long, ignoreRoot: Boolean): Employee {
        return super.getById(id).apply {
            if (ignoreRoot && isRoot) throw EntityNotFoundRestException(id)
        }
    }

    /** Saves the given [employee]. The password contained in the DTO will be hashed in the created [Employee]. */
    fun save(employee: EmployeeDto): Employee {
        return save(with(employee) {
            Employee(id, firstName, lastName, passwordEncoder.encode(password), false, group, permissions, excludedPermissions)
        })
    }

    override fun save(entity: Employee): Employee {
        if (existsByFirstNameAndLastName(entity.firstName, entity.lastName))
            throw EntityAlreadyExistsException(type, ModelException.IdentifierType.NAME, "${entity.firstName} ${entity.lastName}")

        return super.save(entity)
    }

    /** Updates the last login time of the employee with the given [employeeId]. */
    fun updateLastLoginTime(employeeId: Long) {
        update(Employee(id = employeeId, lastLoginTime = LocalDateTime.now()), false)
    }

    override fun update(entity: Employee): Employee {
        return update(entity, true)
    }

    /** Updates de given [entity]. **/
    fun update(entity: Employee, ignoreRoot: Boolean): Employee {
        val persistedEmployee = getById(entity.id, ignoreRoot)
        with(repository.findByFirstNameAndLastName(entity.firstName, entity.lastName)) {
            if (this != null && id != entity.id)
                throw EntityAlreadyExistsRestException("${entity.firstName} ${entity.lastName}")
        }

        return super.update(with(entity) {
            Employee(
                    id,
                    if (firstName.isNotBlank()) firstName else persistedEmployee.firstName,
                    if (lastName.isNotBlank()) lastName else persistedEmployee.lastName,
                    persistedEmployee.password,
                    if (ignoreRoot) false else persistedEmployee.isRoot,
                    persistedEmployee.group,
                    if (permissions.isNotEmpty()) permissions else persistedEmployee.permissions,
                    if (excludedPermissions.isNotEmpty()) excludedPermissions else persistedEmployee.excludedPermissions,
                    lastLoginTime ?: persistedEmployee.lastLoginTime
            )
        })
    }

    /** Adds the given [permission] to the employee with the given [employeeId]. */
    fun addPermission(employeeId: Long, permission: EmployeePermission) = super.update(getById(employeeId).apply { permissions += permission })

    /** Removes the given [permission] from the employee with the given [employeeId]. */
    fun removePermission(employeeId: Long, permission: EmployeePermission) = super.update(getById(employeeId).apply { permissions -= permission })

    /** Adds the given [excludedPermission] to the employee with the given [employeeId]. */
    fun addExcludedPermission(employeeId: Long, excludedPermission: EmployeePermission) = super.update(getById(employeeId).apply { excludedPermissions += excludedPermission })

    /** Removes the given [excludedPermission] to the employee with the given [employeeId]. */
    fun removeExcludedPermission(employeeId: Long, excludedPermission: EmployeePermission) = super.update(getById(employeeId).apply { excludedPermissions -= excludedPermission })
}

@Service
class EmployeeGroupService(val employeeGroupRepository: EmployeeGroupRepository, val employeeService: EmployeeService) : AbstractModelService<EmployeeGroup, EmployeeGroupRepository>(employeeGroupRepository, EmployeeGroup::class.java) {
    /** Adds the employee with the given [employeeId] to the group with the given [groupId]. */
    fun addEmployeeToGroup(groupId: Long, employeeId: Long) {
        addEmployeeToGroup(getById(groupId), employeeService.getById(employeeId))
    }

    /**
     * Adds a given [employee] to a given [group].
     *
     * If the [employee] is already in the [group], nothing will be done.
     * If the [employee] is already in a group, it will be removed from it.
     */
    @Transactional
    fun addEmployeeToGroup(group: EmployeeGroup, employee: Employee) {
        if (employee.group != null) removeEmployeeFromGroup(employee.group!!, employee)
        if (employee.group == group) return

        group.employees.add(employee)
        employee.group = group
        update(group)
        employeeService.update(employee)
    }

    /** Removes the employee with the given [employeeId] from the group with the given [groupId]. */
    fun removeEmployeeFromGroup(groupId: Long, employeeId: Long) =
            removeEmployeeFromGroup(getById(groupId), employeeService.getById(employeeId))

    /** Removes a given [employee] from the given [group]. */
    @Transactional
    fun removeEmployeeFromGroup(group: EmployeeGroup, employee: Employee) {
        if (employee.group == null) return

        group.employees.remove(employee)
        employee.group = null
        update(group)
        employeeService.update(employee)
    }
}

@Service
class EmployeeUserDetailsService(val employeeService: EmployeeService) : UserDetailsService {
    override fun loadUserByUsername(username: String): UserDetails {
        try {
            return loadUserByEmployeeId(username.toLong())
        } catch (ex: EntityNotFoundException) {
            throw UsernameNotFoundException(username)
        }
    }

    /** Loads an [User] for the given [employeeId]. */
    fun loadUserByEmployeeId(employeeId: Long): UserDetails {
        val employee = employeeService.getById(employeeId, false)
        return User(employee.id.toString(), employee.password, employee.getAuthorities())
    }
}
