package dev.fyloz.trial.colorrecipesexplorer.core.services

import dev.fyloz.trial.colorrecipesexplorer.core.exception.TooLowQuantityException
import dev.fyloz.trial.colorrecipesexplorer.core.model.Material
import dev.fyloz.trial.colorrecipesexplorer.core.model.dto.InventoryDto
import dev.fyloz.trial.colorrecipesexplorer.core.services.model.MaterialService
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.stereotype.Service

@Service
class InventoryService {
    @Autowired
    lateinit var materialService: MaterialService

    /**
     * Use all materials in the given [mixes].
     * @throws TooLowQuantityException When there is not enough stock for a material
     */
    fun use(mixes: List<InventoryDto>) {
        mixes.forEach { mix ->
            val materials = mix.materialIds.toMaterials()

            materials.forEachIndexed { index, material ->
                val quantity = mix.quantities[index]

                if (!material.hasEnoughStock(quantity))
                    throw TooLowQuantityException(mix.mixId, material, material.inventoryQuantity, quantity)

                material.use(quantity)
            }
        }
    }

    /** Convert material ids to [Material]s. */
    private fun List<Long>.toMaterials(): List<Material> {
        return map { materialService.getById(it) }
    }

    /** Check if there is at least the given [quantity] in stock for a [Material]. */
    private fun Material.hasEnoughStock(quantity: Float): Boolean {
        return inventoryQuantity >= quantity
    }

    /** Use a given [Material], for the given [quantity]. */
    private fun Material.use(quantity: Float) {
        inventoryQuantity -= quantity
        materialService.update(this)
    }
}
