package dev.fyloz.trial.colorrecipesexplorer.core.services

import com.fasterxml.jackson.module.kotlin.jacksonObjectMapper
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityNotFoundException
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityNotFoundRestException
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.ModelException
import dev.fyloz.trial.colorrecipesexplorer.core.model.IModel
import org.springframework.data.jpa.repository.JpaRepository
import org.springframework.data.repository.findByIdOrNull

/** A service implementing the basics CRUD operations. */
interface IGenericService<E> {
    /** Gets all entities. */
    fun getAll(): Collection<E>

    /** Saves a given [entity]. */
    fun save(entity: E): E

    /** Saves all given [entities]. */
    fun saveAll(entities: Iterable<E>): Collection<E>

    /** Updates a given [entity]. */
    fun update(entity: E): E

    /** Deletes a given [entity]. */
    fun delete(entity: E)

    /** Deletes all give [entities]. */
    fun deleteAll(entities: Iterable<E>)
}

/** A service for entities implementing the [IModel] interface. This service implements CRUD operations for [Long] identifiers. */
interface IGenericModelService<E : IModel> : IGenericService<E> {
    /** Checks if an entity with the given [id] exists. */
    fun existsById(id: Long): Boolean

    /** Gets the entity with the given [id]. */
    fun getById(id: Long): E

    /** Deletes the entity with the given [id]. */
    fun deleteById(id: Long)
}

abstract class AbstractService<E, R : JpaRepository<E, *>>(val repository: R, val type: Class<out IModel>) : IGenericService<E> {
    override fun getAll(): Collection<E> = repository.findAll()

    override fun save(entity: E): E = repository.save(entity)

    override fun saveAll(entities: Iterable<E>): Collection<E> = entities.map(this::save)

    override fun update(entity: E): E = repository.save(entity)

    override fun delete(entity: E) = repository.delete(entity)

    override fun deleteAll(entities: Iterable<E>) = entities.forEach(this::delete)
}

abstract class AbstractModelService<E : IModel, R : JpaRepository<E, Long>>(repository: R, type: Class<out IModel>) : AbstractService<E, R>(repository, type), IGenericModelService<E> {
    override fun existsById(id: Long): Boolean = repository.existsById(id)

    override fun getById(id: Long): E = repository.findByIdOrNull(id)
            ?: throw EntityNotFoundException(type, ModelException.IdentifierType.ID, id)

    override fun save(entity: E): E {
        with(entity.id) {
            if (this != null && existsById(this))
                throw EntityNotFoundRestException(this)
        }

        return super.save(entity)
    }

    override fun deleteById(id: Long) = delete(getById(id))
}

/** Transforms the given object to JSON. **/
fun Any.asJson(): String {
    return jacksonObjectMapper().writeValueAsString(this)
}
