package dev.fyloz.trial.colorrecipesexplorer.rest

import dev.fyloz.trial.colorrecipesexplorer.core.model.Employee
import dev.fyloz.trial.colorrecipesexplorer.core.model.EmployeeDto
import dev.fyloz.trial.colorrecipesexplorer.core.model.EmployeeGroup
import dev.fyloz.trial.colorrecipesexplorer.core.model.EmployeePermission
import dev.fyloz.trial.colorrecipesexplorer.core.services.EmployeeGroupService
import dev.fyloz.trial.colorrecipesexplorer.core.services.EmployeeService
import org.springframework.context.annotation.Profile
import org.springframework.http.HttpStatus
import org.springframework.http.ResponseEntity
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import java.net.URI
import java.security.Principal
import javax.validation.Valid

private const val EMPLOYEE_CONTROLLER_PATH = "api/employee"
private const val EMPLOYEE_GROUP_CONTROLLER_PATH = "api/employee/group"

@RestController
@RequestMapping(EMPLOYEE_CONTROLLER_PATH)
@Profile("rest")
class EmployeeController(employeeService: EmployeeService) :
        AbstractRestModelController<Employee, EmployeeService>(employeeService, EMPLOYEE_CONTROLLER_PATH) {
    @GetMapping("current")
    @ResponseStatus(HttpStatus.OK)
    fun getCurrent(loggedInEmployee: Principal): ResponseEntity<Employee> = getById(loggedInEmployee.name.toLong())

    @PostMapping
    @ResponseStatus(HttpStatus.CREATED)
    fun save(@Valid @RequestBody employee: EmployeeDto): ResponseEntity<Employee> {
        val saved = service.save(employee)
        return ResponseEntity
                .created(URI("$controllerPath/${getEntityId(saved)}"))
                .body(saved)
    }

    @PostMapping("create")
    @ResponseStatus(HttpStatus.NOT_FOUND)
    override fun save(entity: Employee): ResponseEntity<Employee> = ResponseEntity.notFound().build()

    @PutMapping("{employeeId}/permissions/{permission}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("hasAnyAuthority('EDIT_EMPLOYEE')")
    fun addPermission(@PathVariable employeeId: Long, @PathVariable permission: EmployeePermission): ResponseEntity<Void> {
        service.addPermission(employeeId, permission)
        return ResponseEntity
                .noContent()
                .build()
    }

    @DeleteMapping("{employeeId}/permissions/{permission}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("hasAnyAuthority('EDIT_EMPLOYEE')")
    fun removePermission(@PathVariable employeeId: Long, @PathVariable permission: EmployeePermission): ResponseEntity<Void> {
        service.removePermission(employeeId, permission)
        return ResponseEntity
                .noContent()
                .build()
    }

    @PutMapping("{employeeId}/excludedPermissions/{excludedPermission}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("hasAnyAuthority('EDIT_EMPLOYEE')")
    fun addExcludedPermission(@PathVariable employeeId: Long, @PathVariable excludedPermission: EmployeePermission): ResponseEntity<Void> {
        service.addExcludedPermission(employeeId, excludedPermission)
        return ResponseEntity
                .noContent()
                .build()
    }

    @DeleteMapping("{employeeId}/excludedPermissions/{excludedPermission}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    @PreAuthorize("hasAnyAuthority('EDIT_EMPLOYEE')")
    fun removeExcludedPermission(@PathVariable employeeId: Long, @PathVariable excludedPermission: EmployeePermission): ResponseEntity<Void> {
        service.removeExcludedPermission(employeeId, excludedPermission)
        return ResponseEntity
                .noContent()
                .build()
    }
}

@RestController
@RequestMapping(EMPLOYEE_GROUP_CONTROLLER_PATH)
@Profile("rest")
class GroupsController(groupService: EmployeeGroupService) :
        AbstractRestModelController<EmployeeGroup, EmployeeGroupService>(groupService, EMPLOYEE_GROUP_CONTROLLER_PATH) {
    @PutMapping("{groupId}/{employeeId}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    fun addEmployeeToGroup(@PathVariable groupId: Long, @PathVariable employeeId: Long): ResponseEntity<Void> {
        service.addEmployeeToGroup(groupId, employeeId)
        return ResponseEntity
                .noContent()
                .build()
    }

    @DeleteMapping("{groupId}/{employeeId}")
    @ResponseStatus(HttpStatus.NO_CONTENT)
    fun removeEmployeeFromGroup(@PathVariable groupId: Long, @PathVariable employeeId: Long): ResponseEntity<Void> {
        service.removeEmployeeFromGroup(groupId, employeeId)
        return ResponseEntity
                .noContent()
                .build()
    }
}
