package dev.fyloz.trial.colorrecipesexplorer.rest

import dev.fyloz.trial.colorrecipesexplorer.core.model.IModel
import dev.fyloz.trial.colorrecipesexplorer.core.services.IGenericModelService
import dev.fyloz.trial.colorrecipesexplorer.core.services.IGenericService
import org.springframework.http.HttpStatus
import org.springframework.http.ResponseEntity
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import java.net.URI
import javax.validation.Valid

interface IRestController<E> {
    @ResponseStatus(HttpStatus.OK)
    fun getAll(): ResponseEntity<Iterable<E>>

    @ResponseStatus(HttpStatus.CREATED)
    fun save(entity: E): ResponseEntity<E>

    @ResponseStatus(HttpStatus.NO_CONTENT)
    fun update(entity: E): ResponseEntity<Void>

    @ResponseStatus(HttpStatus.NO_CONTENT)
    fun delete(entity: E): ResponseEntity<Void>
}

interface IRestModelController<E : IModel> : IRestController<E> {
    @ResponseStatus(HttpStatus.OK)
    fun getById(id: Long): ResponseEntity<E>

    @ResponseStatus(HttpStatus.NO_CONTENT)
    fun deleteById(id: Long): ResponseEntity<Void>
}

abstract class AbstractRestController<E, S : IGenericService<E>>(val service: S, protected val controllerPath: String) :
        IRestController<E> {
    protected abstract fun getEntityId(entity: E): Any?

    @GetMapping
    override fun getAll(): ResponseEntity<Iterable<E>> = ResponseEntity.ok(service.getAll())

    @PostMapping
    override fun save(@Valid @RequestBody entity: E): ResponseEntity<E> {
        val saved = service.save(entity)
        return ResponseEntity
                .created(URI("$controllerPath/${getEntityId(saved)}"))
                .body(saved)
    }

    @PutMapping
    override fun update(@Valid @RequestBody entity: E): ResponseEntity<Void> {
        service.update(entity)
        return ResponseEntity
                .noContent()
                .build()
    }

    @DeleteMapping
    override fun delete(@Valid @RequestBody entity: E): ResponseEntity<Void> {
        service.delete(entity)
        return ResponseEntity
                .noContent()
                .build()
    }
}

abstract class AbstractRestModelController<E : IModel, S : IGenericModelService<E>>(service: S, controllerPath: String) :
        AbstractRestController<E, S>(service, controllerPath), IRestModelController<E> {
    override fun getEntityId(entity: E) = entity.id

    @GetMapping("{id}")
    override fun getById(@PathVariable id: Long): ResponseEntity<E> = ResponseEntity.ok(service.getById(id))

    @DeleteMapping("{id}")
    override fun deleteById(@PathVariable id: Long): ResponseEntity<Void> {
        service.deleteById(id)
        return ResponseEntity
                .noContent()
                .build()
    }
}
