package dev.fyloz.trial.colorrecipesexplorer.services;

import dev.fyloz.trial.colorrecipesexplorer.builders.MaterialTypeBuilder;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.CannotDeleteDefaultMaterialTypeException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.CannotEditDefaultMaterialTypeException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityAlreadyExistsException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.EntityNotFoundException;
import dev.fyloz.trial.colorrecipesexplorer.core.exception.model.ModelException;
import dev.fyloz.trial.colorrecipesexplorer.core.model.MaterialType;
import dev.fyloz.trial.colorrecipesexplorer.core.model.dto.MaterialTypeEditorDto;
import dev.fyloz.trial.colorrecipesexplorer.core.services.model.MaterialService;
import dev.fyloz.trial.colorrecipesexplorer.core.services.model.MaterialTypeService;
import dev.fyloz.trial.colorrecipesexplorer.dao.MaterialTypeDao;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.Mockito;
import org.mockito.Spy;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.test.context.junit.jupiter.SpringExtension;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

@ExtendWith(SpringExtension.class)
@SpringBootTest
public class MaterialTypeServiceTest extends AbstractServiceTest<MaterialType, MaterialTypeService, MaterialTypeDao> {

    @MockBean
    private MaterialTypeDao materialTypeDao;

    @Autowired
    @SpyBean
    private MaterialTypeService materialTypeService;

    @MockBean
    private MaterialService materialService;

    @Spy
    private List<MaterialType> defaultMaterialTypes = new LinkedList<>();

    @BeforeEach
    void setUp() {
        materialTypeService.setMaterialTypeDao(materialTypeDao);
        materialTypeService.setMaterialService(materialService);
        materialTypeService.setDefaultMaterialTypes(defaultMaterialTypes);

        super.init(materialTypeService, materialTypeDao);

        MaterialType materialType = getEntity();
        MaterialType nameMaterialType = getEntityDifferentName();
        MaterialType prefixMaterialType = getEntityDifferentPrefix();

        doReturn(false).when(defaultMaterialTypes).contains(Mockito.<MaterialType>any());
        doReturn(true).when(defaultMaterialTypes).contains(materialType);

        when(materialTypeDao.existsByName(anyString())).thenReturn(false);
        when(materialTypeDao.existsByName(nameMaterialType.getName())).thenReturn(true);

        when(materialTypeDao.existsByPrefix(anyString())).thenReturn(false);
        when(materialTypeDao.existsByPrefix(prefixMaterialType.getPrefix())).thenReturn(true);

        when(materialService.existsByMaterialType(any())).thenReturn(false);
        when(materialService.existsByMaterialType(materialType)).thenReturn(true);

        when(materialTypeDao.findByName(anyString())).thenReturn(Optional.empty());
        when(materialTypeDao.findByName(nameMaterialType.getName())).thenReturn(Optional.of(nameMaterialType));
    }


    @Test
    public void whenAddDefault_thenListIsCalled() {
        // given
        MaterialType materialType = getEntity();

        // when
        materialTypeService.addDefault(materialType);

        // then
        verify(defaultMaterialTypes).add(materialType);
    }

    @Test
    public void whenIsDefault_thenReturnTrue() {
        // given
        MaterialType materialType = getEntity();

        // when
        boolean found = materialTypeService.isDefault(materialType);

        // then
        assertTrue(found);
    }

    @Test
    public void whenIsDefault_thenReturnFalse() {
        // given
        MaterialType materialType = getEntityDifferentName();

        // when
        boolean found = materialTypeService.isDefault(materialType);

        // then
        assertFalse(found);
    }

    @Test
    public void whenExistsByName_thenReturnTrue() {
        // given
        MaterialType materialType = getEntityDifferentName();

        // when
        boolean found = materialTypeService.existsByName(materialType.getName());

        // then
        assertTrue(found);
    }

    @Test
    public void whenExistsByName_thenReturnFalse() {
        // given
        MaterialType materialType = getEntity();

        // when
        boolean found = materialTypeService.existsByName(materialType.getName());

        // then
        assertFalse(found);
    }

    @Test
    public void whenExistsByPrefix_thenReturnTrue() {
        // given
        MaterialType materialType = getEntityDifferentPrefix();

        // when
        boolean found = materialTypeService.existsByPrefix(materialType.getPrefix());

        // then
        assertTrue(found);
    }

    @Test
    public void whenExistsByPrefix_thenReturnFalse() {
        // given
        MaterialType materialType = getEntity();

        // when
        boolean found = materialTypeService.existsByPrefix(materialType.getPrefix());

        // then
        assertFalse(found);
    }

    @Test
    public void whenIsLinkedToMaterials_thenReturnTrue() {
        // given
        MaterialType materialType = getEntity();

        // when
        boolean found = materialTypeService.isLinkedToMaterials(materialType);

        // then
        assertTrue(found);
    }

    @Test
    public void whenIsLinkedToMaterials_thenReturnFalse() {
        // given
        MaterialType materialType = getEntityDifferentName();

        // when
        boolean found = materialTypeService.isLinkedToMaterials(materialType);

        // then
        assertFalse(found);
    }

    @Test
    public void whenGetAllNotDefault_thenReturnMaterialTypeList() {
        // given
        MaterialType defaultMaterialType = getEntity();
        MaterialType materialType = getEntityDifferentName();
        List<MaterialType> materialTypes = Arrays.asList(defaultMaterialType, materialType);

        doReturn(materialTypes).when(materialTypeService).getAll();

        // when
        List<MaterialType> found = materialTypeService.getAllNotDefault();

        // then
        assertFalse(found.isEmpty());
        assertTrue(found.contains(materialType));
        assertFalse(found.contains(defaultMaterialType));
    }

    @Test
    public void whenGetByName_thenReturnTrue() {
        // given
        MaterialType materialType = getEntityDifferentName();

        // when
        MaterialType found = materialTypeService.getByName(materialType.getName());

        // then
        assertEquals(materialType, found);
    }

    @Test
    public void whenGetByName_thenThrowEntityNotFoundExceptionName() {
        // given
        MaterialType materialType = getEntity();

        // then
        EntityNotFoundException thrown = assertThrows(EntityNotFoundException.class, () -> materialTypeService.getByName(materialType.getName()));
        assertEquals(thrown.getIdentifierType(), ModelException.IdentifierType.NAME);
    }

    @Test
    public void whenUpdate_thenReturnMaterialType() {
        // given
        MaterialType materialType = getEntityDifferentName();
        MaterialTypeEditorDto dto = new MaterialTypeEditorDto(materialType);

        doReturn(false).when(materialTypeService).isDefault(materialType);

        // when
        MaterialType found = materialTypeService.update(dto);

        // then
        assertEquals(materialType, found);
    }

    @Test
    public void whenUpdate_thenThrowCannotEditDefaultMaterialTypeException() {
        // given
        MaterialType materialType = getEntity();

        // then
        assertThrows(CannotEditDefaultMaterialTypeException.class, () -> materialTypeService.update(new MaterialTypeEditorDto(materialType)));
    }

    @Test
    public void whenUpdate_thenThrowEntityNotFoundExceptionOldName() {
        // given
        MaterialType materialType = getEntityDifferentPrefix();

        // then
        EntityNotFoundException thrown = assertThrows(EntityNotFoundException.class, () -> materialTypeService.update(new MaterialTypeEditorDto(materialType)));
        assertEquals(thrown.getIdentifierType(), ModelException.IdentifierType.NAME);
    }

    @Test
    public void whenUpdate_thenThrowEntityAlreadyExistsExceptionName() {
        // given
        MaterialType materialType = getEntityDifferentName();
        MaterialTypeEditorDto dto = new MaterialTypeEditorDto(materialType);
        dto.setOldName("Base");

        doReturn(false).when(materialTypeService).isDefault(dto.getOldMaterialType());
        doReturn(true).when(materialTypeService).existsByName(dto.getOldName());

        // then
        EntityAlreadyExistsException thrown = assertThrows(EntityAlreadyExistsException.class, () -> materialTypeService.update(dto));
        assertEquals(thrown.getIdentifierType(), ModelException.IdentifierType.NAME);
    }

    @Test
    public void whenUpdate_thenThrowEntityAlreadyExistsExceptionPrefix() {
        // given
        MaterialType materialType = getEntityDifferentPrefix();
        MaterialTypeEditorDto dto = new MaterialTypeEditorDto(materialType);
        dto.setOldPrefix("BAS");
        dto.setName("Catalyseur");
        dto.setOldName("Catalyseur");

        // then
        EntityAlreadyExistsException thrown = assertThrows(EntityAlreadyExistsException.class, () -> materialTypeService.update(dto));
        assertEquals(thrown.getIdentifierType(), ModelException.IdentifierType.OTHER);
        assertEquals(thrown.getIdentifierName(), MaterialType.IDENTIFIER_PREFIX_NAME);
    }

    @Override
    public void whenDelete_thenDaoIsCalled() {
        // given
        MaterialType materialType = getEntityDifferentName();

        // when
        materialTypeService.delete(materialType);

        // then
        verify(materialTypeDao).delete(materialType);
    }

    @Test
    public void whenDelete_thenThrowCannotDeleteDefaultMaterialTypeException() {
        // given
        MaterialType materialType = getEntity();

        // then
        assertThrows(CannotDeleteDefaultMaterialTypeException.class, () -> materialTypeService.delete(materialType));
    }

    @Override
    public void whenDeleteById_thenDeleteIsCalled() {
        // given
        MaterialType materialType = getEntityDifferentId();

        // when
        materialTypeService.deleteById(materialType.getId());

        // then
        verify(materialTypeService).delete(materialType);
    }

    @Override
    protected MaterialType getEntity() {
        return new MaterialTypeBuilder().build();
    }

    @Override
    protected MaterialType getEntityDifferentId() {
        return new MaterialTypeBuilder()
                .id(1L)
                .build();
    }

    @Override
    protected MaterialType getEntityNullId() {
        return new MaterialTypeBuilder()
                .id(null)
                .build();
    }

    private MaterialType getEntityDifferentName() {
        return new MaterialTypeBuilder()
                .name("Catalyseur")
                .build();
    }

    private MaterialType getEntityDifferentPrefix() {
        return new MaterialTypeBuilder()
                .prefix("CAT")
                .build();
    }
}
